@@ -1,7 +1,7 @@
 /*
    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    SLEPc - Scalable Library for Eigenvalue Problem Computations
-   Copyright (c) 2002-2018, Universitat Politecnica de Valencia, Spain
+   Copyright (c) 2002-2021, Universitat Politecnica de Valencia, Spain
 
    This file is part of SLEPc.
    SLEPc is distributed under a 2-clause BSD license (see LICENSE).
@@ -22,61 +22,20 @@ PETSC_STATIC_INLINE PetscErrorCode LMESetUp_Lyapunov(LME lme)
   PetscFunctionBegin;
   ierr = MatLRCGetMats(lme->C,NULL,&C1,&dc,&C2);CHKERRQ(ierr);
   if (C1!=C2) SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_ARG_WRONGSTATE,"Lyapunov matrix equation requires symmetric right-hand side C");
-  if (dc) SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_ARG_WRONGSTATE,"Lyapunov solvers currently requires positive-definite right-hand side C");
+  if (dc) SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_ARG_WRONGSTATE,"Lyapunov solvers currently require positive-definite right-hand side C");
   if (lme->X) {
     ierr = MatLRCGetMats(lme->X,NULL,&X1,&dx,&X2);CHKERRQ(ierr);
     if (X1!=X2) SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_ARG_WRONGSTATE,"Lyapunov matrix equation requires symmetric solution X");
-    if (dx) SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_ARG_WRONGSTATE,"Lyapunov solvers currently assumes a positive-definite solution X");
+    if (dx) SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_ARG_WRONGSTATE,"Lyapunov solvers currently assume a positive-definite solution X");
   }
   PetscFunctionReturn(0);
 }
 
-PETSC_STATIC_INLINE PetscErrorCode LMESetUp_Sylvester(LME lme)
-{
-  PetscFunctionBegin;
-  if (!lme->B) SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_ARG_WRONGSTATE,"Sylvester matrix equation requires coefficient matrix B");
-  SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_SUP,"There is no solver yet for this matrix equation type");
-  PetscFunctionReturn(0);
-}
-
-PETSC_STATIC_INLINE PetscErrorCode LMESetUp_Gen_Lyapunov(LME lme)
-{
-  PetscFunctionBegin;
-  if (!lme->D) SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_ARG_WRONGSTATE,"Generalized Lyapunov matrix equation requires coefficient matrix D");
-  SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_SUP,"There is no solver yet for this matrix equation type");
-  PetscFunctionReturn(0);
-}
-
-PETSC_STATIC_INLINE PetscErrorCode LMESetUp_Gen_Sylvester(LME lme)
-{
-  PetscFunctionBegin;
-  if (!lme->B) SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_ARG_WRONGSTATE,"Generalized Sylvester matrix equation requires coefficient matrix B");
-  if (!lme->D) SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_ARG_WRONGSTATE,"Generalized Sylvester matrix equation requires coefficient matrix D");
-  if (!lme->E) SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_ARG_WRONGSTATE,"Generalized Sylvester matrix equation requires coefficient matrix E");
-  SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_SUP,"There is no solver yet for this matrix equation type");
-  PetscFunctionReturn(0);
-}
-
-PETSC_STATIC_INLINE PetscErrorCode LMESetUp_Stein(LME lme)
-{
-  PetscFunctionBegin;
-  if (!lme->D) SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_ARG_WRONGSTATE,"Stein matrix equation requires coefficient matrix D");
-  SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_SUP,"There is no solver yet for this matrix equation type");
-  PetscFunctionReturn(0);
-}
-
-PETSC_STATIC_INLINE PetscErrorCode LMESetUp_DT_Lyapunov(LME lme)
-{
-  PetscFunctionBegin;
-  SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_SUP,"There is no solver yet for this matrix equation type");
-  PetscFunctionReturn(0);
-}
-
 /*@
    LMESetUp - Sets up all the internal data structures necessary for the
    execution of the linear matrix equation solver.
 
-   Collective on LME
+   Collective on lme
 
    Input Parameter:
 .  lme   - linear matrix equation solver context
@@ -120,21 +79,23 @@ PetscErrorCode LMESetUp(LME lme)
       ierr = LMESetUp_Lyapunov(lme);CHKERRQ(ierr);
       break;
     case LME_SYLVESTER:
-      ierr = LMESetUp_Sylvester(lme);CHKERRQ(ierr);
+      LMECheckCoeff(lme,lme->B,"B","Sylvester");
       break;
     case LME_GEN_LYAPUNOV:
-      ierr = LMESetUp_Gen_Lyapunov(lme);CHKERRQ(ierr);
+      LMECheckCoeff(lme,lme->D,"D","Generalized Lyapunov");
       break;
     case LME_GEN_SYLVESTER:
-      ierr = LMESetUp_Gen_Sylvester(lme);CHKERRQ(ierr);
+      LMECheckCoeff(lme,lme->B,"B","Generalized Sylvester");
+      LMECheckCoeff(lme,lme->D,"D","Generalized Sylvester");
+      LMECheckCoeff(lme,lme->E,"E","Generalized Sylvester");
       break;
     case LME_DT_LYAPUNOV:
-      ierr = LMESetUp_DT_Lyapunov(lme);CHKERRQ(ierr);
       break;
     case LME_STEIN:
-      ierr = LMESetUp_Stein(lme);CHKERRQ(ierr);
+      LMECheckCoeff(lme,lme->D,"D","Stein");
       break;
   }
+  if (lme->problem_type!=LME_LYAPUNOV) SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_SUP,"There is no solver yet for this matrix equation type");
 
   /* call specific solver setup */
   ierr = (*lme->ops->setup)(lme);CHKERRQ(ierr);
@@ -147,11 +108,25 @@ PetscErrorCode LMESetUp(LME lme)
   PetscFunctionReturn(0);
 }
 
+PETSC_STATIC_INLINE PetscErrorCode LMESetCoefficients_Private(LME lme,Mat A,Mat *lmeA)
+{
+  PetscErrorCode ierr;
+  PetscInt       m,n;
+
+  PetscFunctionBegin;
+  ierr = MatGetSize(A,&m,&n);CHKERRQ(ierr);
+  if (m!=n) SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_ARG_WRONG,"Matrix is non-square");
+  if (!lme->setupcalled) { ierr = MatDestroy(lmeA);CHKERRQ(ierr); }
+  ierr = PetscObjectReference((PetscObject)A);CHKERRQ(ierr);
+  *lmeA = A;
+  PetscFunctionReturn(0);
+}
+
 /*@
    LMESetCoefficients - Sets the coefficient matrices that define the linear matrix
    equation to be solved.
 
-   Collective on LME and Mat
+   Collective on lme
 
    Input Parameters:
 +  lme - the matrix function context
@@ -178,7 +153,6 @@ PetscErrorCode LMESetUp(LME lme)
 PetscErrorCode LMESetCoefficients(LME lme,Mat A,Mat B,Mat D,Mat E)
 {
   PetscErrorCode ierr;
-  PetscInt       m,n;
 
   PetscFunctionBegin;
   PetscValidHeaderSpecific(lme,LME_CLASSID,1);
@@ -199,32 +173,13 @@ PetscErrorCode LMESetCoefficients(LME lme,Mat A,Mat B,Mat D,Mat E)
 
   if (lme->setupcalled) { ierr = LMEReset(lme);CHKERRQ(ierr); }
 
-  ierr = MatGetSize(A,&m,&n);CHKERRQ(ierr);
-  if (m!=n) SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_ARG_WRONG,"A is a non-square matrix");
-  if (!lme->setupcalled) { ierr = MatDestroy(&lme->A);CHKERRQ(ierr); }
-  ierr = PetscObjectReference((PetscObject)A);CHKERRQ(ierr);
-  lme->A = A;
-  if (B) {
-    ierr = MatGetSize(B,&m,&n);CHKERRQ(ierr);
-    if (m!=n) SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_ARG_WRONG,"B is a non-square matrix");
-    if (!lme->setupcalled) { ierr = MatDestroy(&lme->B);CHKERRQ(ierr); }
-    ierr = PetscObjectReference((PetscObject)B);CHKERRQ(ierr);
-    lme->B = B;
-  } else if (!lme->setupcalled) { ierr = MatDestroy(&lme->B);CHKERRQ(ierr); }
-  if (D) {
-    ierr = MatGetSize(D,&m,&n);CHKERRQ(ierr);
-    if (m!=n) SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_ARG_WRONG,"D is a non-square matrix");
-    if (!lme->setupcalled) { ierr = MatDestroy(&lme->D);CHKERRQ(ierr); }
-    ierr = PetscObjectReference((PetscObject)D);CHKERRQ(ierr);
-    lme->D = D;
-  } else if (!lme->setupcalled) { ierr = MatDestroy(&lme->D);CHKERRQ(ierr); }
-  if (E) {
-    ierr = MatGetSize(E,&m,&n);CHKERRQ(ierr);
-    if (m!=n) SETERRQ(PetscObjectComm((PetscObject)lme),PETSC_ERR_ARG_WRONG,"E is a non-square matrix");
-    if (!lme->setupcalled) { ierr = MatDestroy(&lme->E);CHKERRQ(ierr); }
-    ierr = PetscObjectReference((PetscObject)E);CHKERRQ(ierr);
-    lme->E = E;
-  } else if (!lme->setupcalled) { ierr = MatDestroy(&lme->E);CHKERRQ(ierr); }
+  ierr = LMESetCoefficients_Private(lme,A,&lme->A);CHKERRQ(ierr);
+  if (B) { ierr = LMESetCoefficients_Private(lme,B,&lme->B);CHKERRQ(ierr); }
+  else if (!lme->setupcalled) { ierr = MatDestroy(&lme->B);CHKERRQ(ierr); }
+  if (D) { ierr = LMESetCoefficients_Private(lme,D,&lme->D);CHKERRQ(ierr); }
+  else if (!lme->setupcalled) { ierr = MatDestroy(&lme->D);CHKERRQ(ierr); }
+  if (E) { ierr = LMESetCoefficients_Private(lme,E,&lme->E);CHKERRQ(ierr); }
+  else if (!lme->setupcalled) { ierr = MatDestroy(&lme->E);CHKERRQ(ierr); }
 
   lme->setupcalled = 0;
   PetscFunctionReturn(0);
@@ -233,7 +188,7 @@ PetscErrorCode LMESetCoefficients(LME lme,Mat A,Mat B,Mat D,Mat E)
 /*@
    LMEGetCoefficients - Gets the coefficient matrices of the matrix equation.
 
-   Collective on LME and Mat
+   Collective on lme
 
    Input Parameter:
 .  lme - the LME context
@@ -263,7 +218,7 @@ PetscErrorCode LMEGetCoefficients(LME lme,Mat *A,Mat *B,Mat *D,Mat *E)
    LMESetRHS - Sets the right-hand side of the matrix equation, as a low-rank
    matrix.
 
-   Collective on LME and Mat
+   Collective on lme
 
    Input Parameters:
 +  lme - the matrix function context
@@ -286,21 +241,19 @@ PetscErrorCode LMEGetCoefficients(LME lme,Mat *A,Mat *B,Mat *D,Mat *E)
 PetscErrorCode LMESetRHS(LME lme,Mat C)
 {
   PetscErrorCode ierr;
-  PetscBool      match;
   Mat            A;
 
   PetscFunctionBegin;
   PetscValidHeaderSpecific(lme,LME_CLASSID,1);
   PetscValidHeaderSpecific(C,MAT_CLASSID,2);
   PetscCheckSameComm(lme,1,C,2);
+  PetscCheckTypeName(C,MATLRC);
 
-  ierr = PetscObjectTypeCompare((PetscObject)C,MATLRC,&match);CHKERRQ(ierr);
-  if (!match) SETERRQ(PetscObjectComm((PetscObject)C),PETSC_ERR_SUP,"Mat argument must have been created with MatCreateLRC");
   ierr = MatLRCGetMats(C,&A,NULL,NULL,NULL);CHKERRQ(ierr);
   if (A) SETERRQ(PetscObjectComm((PetscObject)C),PETSC_ERR_SUP,"The MatLRC must not have a sparse matrix term");
 
-  ierr = MatDestroy(&lme->C);CHKERRQ(ierr);
   ierr = PetscObjectReference((PetscObject)C);CHKERRQ(ierr);
+  ierr = MatDestroy(&lme->C);CHKERRQ(ierr);
   lme->C = C;
   PetscFunctionReturn(0);
 }
@@ -308,7 +261,7 @@ PetscErrorCode LMESetRHS(LME lme,Mat C)
 /*@
    LMEGetRHS - Gets the right-hand side of the matrix equation.
 
-   Collective on LME and Mat
+   Collective on lme
 
    Input Parameter:
 .  lme - the LME context
@@ -333,7 +286,7 @@ PetscErrorCode LMEGetRHS(LME lme,Mat *C)
    LMESetSolution - Sets the placeholder for the solution of the matrix
    equation, as a low-rank matrix.
 
-   Collective on LME and Mat
+   Collective on lme
 
    Input Parameters:
 +  lme - the matrix function context
@@ -360,7 +313,6 @@ PetscErrorCode LMEGetRHS(LME lme,Mat *C)
 PetscErrorCode LMESetSolution(LME lme,Mat X)
 {
   PetscErrorCode ierr;
-  PetscBool      match;
   Mat            A;
 
   PetscFunctionBegin;
@@ -368,24 +320,20 @@ PetscErrorCode LMESetSolution(LME lme,Mat X)
   if (X) {
     PetscValidHeaderSpecific(X,MAT_CLASSID,2);
     PetscCheckSameComm(lme,1,X,2);
-    ierr = PetscObjectTypeCompare((PetscObject)X,MATLRC,&match);CHKERRQ(ierr);
-    if (!match) SETERRQ(PetscObjectComm((PetscObject)X),PETSC_ERR_SUP,"Mat argument must have been created with MatCreateLRC");
+    PetscCheckTypeName(X,MATLRC);
     ierr = MatLRCGetMats(X,&A,NULL,NULL,NULL);CHKERRQ(ierr);
     if (A) SETERRQ(PetscObjectComm((PetscObject)X),PETSC_ERR_SUP,"The MatLRC must not have a sparse matrix term");
-  }
-
-  ierr = MatDestroy(&lme->X);CHKERRQ(ierr);
-  if (X) {
     ierr = PetscObjectReference((PetscObject)X);CHKERRQ(ierr);
-    lme->X = X;
   }
+  ierr = MatDestroy(&lme->X);CHKERRQ(ierr);
+  lme->X = X;
   PetscFunctionReturn(0);
 }
 
 /*@
    LMEGetSolution - Gets the solution of the matrix equation.
 
-   Collective on LME and Mat
+   Collective on lme
 
    Input Parameter:
 .  lme - the LME context
@@ -410,7 +358,7 @@ PetscErrorCode LMEGetSolution(LME lme,Mat *X)
    LMEAllocateSolution - Allocate memory storage for common variables such
    as the basis vectors.
 
-   Collective on LME
+   Collective on lme
 
    Input Parameters:
 +  lme   - linear matrix equation solver context
@@ -418,7 +366,7 @@ PetscErrorCode LMEGetSolution(LME lme,Mat *X)
            working basis slightly larger than ncv
 
    Developers Note:
-   This is PETSC_EXTERN because it may be required by user plugin LME
+   This is SLEPC_EXTERN because it may be required by user plugin LME
    implementations.
 
    Level: developer
