@@ -1,7 +1,7 @@
 /*
    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    SLEPc - Scalable Library for Eigenvalue Problem Computations
-   Copyright (c) 2002-2018, Universitat Politecnica de Valencia, Spain
+   Copyright (c) 2002-2021, Universitat Politecnica de Valencia, Spain
 
    This file is part of SLEPc.
    SLEPc is distributed under a 2-clause BSD license (see LICENSE).
@@ -19,7 +19,7 @@ PetscFunctionList BVList = 0;
 /*@C
    BVSetType - Selects the type for the BV object.
 
-   Logically Collective on BV
+   Logically Collective on bv
 
    Input Parameter:
 +  bv   - the basis vectors context
@@ -90,7 +90,7 @@ PetscErrorCode BVGetType(BV bv,BVType *type)
 /*@
    BVSetSizes - Sets the local and global sizes, and the number of columns.
 
-   Collective on BV
+   Collective on bv
 
    Input Parameters:
 +  bv - the basis vectors
@@ -149,7 +149,7 @@ PetscErrorCode BVSetSizes(BV bv,PetscInt n,PetscInt N,PetscInt m)
    BVSetSizesFromVec - Sets the local and global sizes, and the number of columns.
    Local and global sizes are specified indirectly by passing a template vector.
 
-   Collective on BV
+   Collective on bv
 
    Input Parameters:
 +  bv - the basis vectors
@@ -233,7 +233,7 @@ PetscErrorCode BVGetSizes(BV bv,PetscInt *n,PetscInt *N,PetscInt *m)
 /*@
    BVSetNumConstraints - Set the number of constraints.
 
-   Logically Collective on BV
+   Logically Collective on V
 
    Input Parameters:
 +  V  - basis vectors
@@ -281,9 +281,9 @@ PetscErrorCode BVSetNumConstraints(BV V,PetscInt nc)
   V->nc = nc;
   V->ci[0] = -V->nc-1;
   V->ci[1] = -V->nc-1;
-  V->l = 0;
   V->m = total-nc;
-  V->k = V->m;
+  V->l = PetscMin(V->l,V->m);
+  V->k = PetscMin(V->k,V->m);
   ierr = PetscObjectStateIncrease((PetscObject)V);CHKERRQ(ierr);
   PetscFunctionReturn(0);
 }
@@ -307,7 +307,7 @@ PetscErrorCode BVGetNumConstraints(BV bv,PetscInt *nc)
 {
   PetscFunctionBegin;
   PetscValidHeaderSpecific(bv,BV_CLASSID,1);
-  PetscValidPointer(nc,2);
+  PetscValidIntPointer(nc,2);
   *nc = bv->nc;
   PetscFunctionReturn(0);
 }
@@ -315,7 +315,7 @@ PetscErrorCode BVGetNumConstraints(BV bv,PetscInt *nc)
 /*@
    BVResize - Change the number of columns.
 
-   Collective on BV
+   Collective on bv
 
    Input Parameters:
 +  bv   - the basis vectors
@@ -345,6 +345,7 @@ PetscErrorCode BVResize(BV bv,PetscInt m,PetscBool copy)
   if (m <= 0) SETERRQ1(PetscObjectComm((PetscObject)bv),PETSC_ERR_ARG_INCOMP,"Number of columns %D must be positive",m);
   if (bv->nc && !bv->issplit) SETERRQ(PetscObjectComm((PetscObject)bv),PETSC_ERR_ARG_WRONGSTATE,"Cannot resize a BV with constraints");
   if (bv->m == m) PetscFunctionReturn(0);
+  BVCheckOp(bv,1,resize);
 
   ierr = PetscLogEventBegin(BV_Create,bv,0,0,0);CHKERRQ(ierr);
   ierr = (*bv->ops->resize)(bv,m,copy);CHKERRQ(ierr);
@@ -353,7 +354,7 @@ PetscErrorCode BVResize(BV bv,PetscInt m,PetscBool copy)
   ierr = PetscFree2(bv->h,bv->c);CHKERRQ(ierr);
   if (bv->omega) {
     if (bv->cuda) {
-#if defined(PETSC_HAVE_VECCUDA)
+#if defined(PETSC_HAVE_CUDA)
       ierr = VecCreateSeqCUDA(PETSC_COMM_SELF,m,&v);CHKERRQ(ierr);
 #else
       SETERRQ(PetscObjectComm((PetscObject)bv),1,"Something wrong happened");
@@ -365,7 +366,7 @@ PetscErrorCode BVResize(BV bv,PetscInt m,PetscBool copy)
     if (copy) {
       ierr = VecGetArray(v,&array);CHKERRQ(ierr);
       ierr = VecGetArrayRead(bv->omega,&omega);CHKERRQ(ierr);
-      ierr = PetscMemcpy(array,omega,PetscMin(m,bv->m)*sizeof(PetscScalar));CHKERRQ(ierr);
+      ierr = PetscArraycpy(array,omega,PetscMin(m,bv->m));CHKERRQ(ierr);
       ierr = VecRestoreArrayRead(bv->omega,&omega);CHKERRQ(ierr);
       ierr = VecRestoreArray(v,&array);CHKERRQ(ierr);
     } else {
@@ -385,7 +386,7 @@ PetscErrorCode BVResize(BV bv,PetscInt m,PetscBool copy)
 /*@
    BVSetActiveColumns - Specify the columns that will be involved in operations.
 
-   Logically Collective on BV
+   Logically Collective on bv
 
    Input Parameters:
 +  bv - the basis vectors context
@@ -459,7 +460,7 @@ PetscErrorCode BVGetActiveColumns(BV bv,PetscInt *l,PetscInt *k)
 /*@
    BVSetMatrix - Specifies the inner product to be used in orthogonalization.
 
-   Collective on BV
+   Collective on bv
 
    Input Parameters:
 +  bv    - the basis vectors context
@@ -483,7 +484,7 @@ PetscErrorCode BVGetActiveColumns(BV bv,PetscInt *l,PetscInt *k)
 
    Level: advanced
 
-.seealso: BVGetMatrix(), BVDot(), BVNorm(), BVOrthogonalize()
+.seealso: BVGetMatrix(), BVDot(), BVNorm(), BVOrthogonalize(), BVSetDefiniteTolerance()
 @*/
 PetscErrorCode BVSetMatrix(BV bv,Mat B,PetscBool indef)
 {
@@ -493,17 +494,21 @@ PetscErrorCode BVSetMatrix(BV bv,Mat B,PetscBool indef)
   PetscFunctionBegin;
   PetscValidHeaderSpecific(bv,BV_CLASSID,1);
   PetscValidLogicalCollectiveBool(bv,indef,3);
-  if (B) {
-    PetscValidHeaderSpecific(B,MAT_CLASSID,2);
-    ierr = MatGetLocalSize(B,&m,&n);CHKERRQ(ierr);
-    if (m!=n) SETERRQ(PetscObjectComm((PetscObject)bv),PETSC_ERR_ARG_SIZ,"Matrix must be square");
-    if (bv->m && bv->n!=n) SETERRQ2(PetscObjectComm((PetscObject)bv),PETSC_ERR_ARG_INCOMP,"Mismatching local dimension BV %D, Mat %D",bv->n,n);
+  if (B!=bv->matrix || (B && ((PetscObject)B)->id!=((PetscObject)bv->matrix)->id) || indef!=bv->indef) {
+    if (B) {
+      PetscValidHeaderSpecific(B,MAT_CLASSID,2);
+      ierr = MatGetLocalSize(B,&m,&n);CHKERRQ(ierr);
+      if (m!=n) SETERRQ(PetscObjectComm((PetscObject)bv),PETSC_ERR_ARG_SIZ,"Matrix must be square");
+      if (bv->m && bv->n!=n) SETERRQ2(PetscObjectComm((PetscObject)bv),PETSC_ERR_ARG_INCOMP,"Mismatching local dimension BV %D, Mat %D",bv->n,n);
+    }
+    if (B) { ierr = PetscObjectReference((PetscObject)B);CHKERRQ(ierr); }
+    ierr = MatDestroy(&bv->matrix);CHKERRQ(ierr);
+    bv->matrix = B;
+    bv->indef  = indef;
+    ierr = PetscObjectStateIncrease((PetscObject)bv);CHKERRQ(ierr);
+    if (bv->Bx) { ierr = PetscObjectStateIncrease((PetscObject)bv->Bx);CHKERRQ(ierr); }
+    if (bv->cached) { ierr = PetscObjectStateIncrease((PetscObject)bv->cached);CHKERRQ(ierr); }
   }
-  ierr = MatDestroy(&bv->matrix);CHKERRQ(ierr);
-  if (B) PetscObjectReference((PetscObject)B);
-  bv->matrix = B;
-  bv->indef  = indef;
-  ierr = PetscObjectStateIncrease((PetscObject)bv);CHKERRQ(ierr);
   PetscFunctionReturn(0);
 }
 
@@ -536,7 +541,7 @@ PetscErrorCode BVGetMatrix(BV bv,Mat *B,PetscBool *indef)
    BVApplyMatrix - Multiplies a vector by the matrix representation of the
    inner product.
 
-   Neighbor-wise Collective on BV and Vec
+   Neighbor-wise Collective on bv
 
    Input Parameter:
 +  bv - the basis vectors context
@@ -573,7 +578,7 @@ PetscErrorCode BVApplyMatrix(BV bv,Vec x,Vec y)
    BVApplyMatrixBV - Multiplies the BV vectors by the matrix representation
    of the inner product.
 
-   Neighbor-wise Collective on BV
+   Neighbor-wise Collective on X
 
    Input Parameter:
 .  X - the basis vectors context
@@ -614,7 +619,7 @@ PetscErrorCode BVApplyMatrixBV(BV X,BV Y)
 /*@
    BVSetSignature - Sets the signature matrix to be used in orthogonalization.
 
-   Logically Collective on BV
+   Logically Collective on bv
 
    Input Parameter:
 +  bv    - the basis vectors context
@@ -705,7 +710,7 @@ PetscErrorCode BVGetSignature(BV bv,Vec omega)
    BVSetBufferVec - Attach a vector object to be used as buffer space for
    several operations.
 
-   Collective on BV
+   Collective on bv
 
    Input Parameters:
 +  bv     - the basis vectors context)
@@ -735,9 +740,9 @@ PetscErrorCode BVSetBufferVec(BV bv,Vec buffer)
   ierr = VecGetSize(buffer,&n);CHKERRQ(ierr);
   ld = bv->m+bv->nc;
   if (n != ld*bv->m) SETERRQ1(PetscObjectComm((PetscObject)bv),PETSC_ERR_ARG_SIZ,"Buffer size must be %d",ld*bv->m);
-  ierr = MPI_Comm_size(PetscObjectComm((PetscObject)buffer),&size);CHKERRQ(ierr);
+  ierr = MPI_Comm_size(PetscObjectComm((PetscObject)buffer),&size);CHKERRMPI(ierr);
   if (size>1) SETERRQ(PetscObjectComm((PetscObject)bv),PETSC_ERR_ARG_WRONG,"Buffer must be a sequential vector");
-  
+
   ierr = PetscObjectReference((PetscObject)buffer);CHKERRQ(ierr);
   ierr = VecDestroy(&bv->buffer);CHKERRQ(ierr);
   bv->buffer = buffer;
@@ -748,7 +753,7 @@ PetscErrorCode BVSetBufferVec(BV bv,Vec buffer)
 /*@
    BVGetBufferVec - Obtain the buffer vector associated with the BV object.
 
-   Not Collective
+   Not Collective, but Vec returned is parallel if BV is parallel
 
    Input Parameters:
 .  bv - the basis vectors context
@@ -803,7 +808,7 @@ PetscErrorCode BVGetBufferVec(BV bv,Vec *buffer)
    BVSetRandomContext - Sets the PetscRandom object associated with the BV,
    to be used in operations that need random numbers.
 
-   Collective on BV
+   Collective on bv
 
    Input Parameters:
 +  bv   - the basis vectors context
@@ -811,7 +816,7 @@ PetscErrorCode BVGetBufferVec(BV bv,Vec *buffer)
 
    Level: advanced
 
-.seealso: BVGetRandomContext(), BVSetRandom(), BVSetRandomColumn()
+.seealso: BVGetRandomContext(), BVSetRandom(), BVSetRandomNormal(), BVSetRandomColumn(), BVSetRandomCond()
 @*/
 PetscErrorCode BVSetRandomContext(BV bv,PetscRandom rand)
 {
@@ -841,7 +846,7 @@ PetscErrorCode BVSetRandomContext(BV bv,PetscRandom rand)
 
    Level: advanced
 
-.seealso: BVSetRandomContext(), BVSetRandom(), BVSetRandomColumn()
+.seealso: BVSetRandomContext(), BVSetRandom(), BVSetRandomNormal(), BVSetRandomColumn(), BVSetRandomCond()
 @*/
 PetscErrorCode BVGetRandomContext(BV bv,PetscRandom* rand)
 {
@@ -853,6 +858,8 @@ PetscErrorCode BVGetRandomContext(BV bv,PetscRandom* rand)
   if (!bv->rand) {
     ierr = PetscRandomCreate(PetscObjectComm((PetscObject)bv),&bv->rand);CHKERRQ(ierr);
     ierr = PetscLogObjectParent((PetscObject)bv,(PetscObject)bv->rand);CHKERRQ(ierr);
+    if (bv->cuda) { ierr = PetscRandomSetType(bv->rand,PETSCCURAND);CHKERRQ(ierr); }
+    if (bv->sfocalled) { ierr = PetscRandomSetFromOptions(bv->rand);CHKERRQ(ierr); }
     if (bv->rrandom) {
       ierr = PetscRandomSetSeed(bv->rand,0x12345678);CHKERRQ(ierr);
       ierr = PetscRandomSeed(bv->rand);CHKERRQ(ierr);
@@ -865,7 +872,7 @@ PetscErrorCode BVGetRandomContext(BV bv,PetscRandom* rand)
 /*@
    BVSetFromOptions - Sets BV options from the options database.
 
-   Collective on BV
+   Collective on bv
 
    Input Parameter:
 .  bv - the basis vectors context
@@ -886,7 +893,7 @@ PetscErrorCode BVSetFromOptions(BV bv)
   PetscValidHeaderSpecific(bv,BV_CLASSID,1);
   ierr = BVRegisterAll();CHKERRQ(ierr);
   ierr = PetscObjectOptionsBegin((PetscObject)bv);CHKERRQ(ierr);
-    ierr = PetscOptionsFList("-bv_type","Basis Vectors type","BVSetType",BVList,(char*)(((PetscObject)bv)->type_name?((PetscObject)bv)->type_name:BVSVEC),type,256,&flg1);CHKERRQ(ierr);
+    ierr = PetscOptionsFList("-bv_type","Basis Vectors type","BVSetType",BVList,(char*)(((PetscObject)bv)->type_name?((PetscObject)bv)->type_name:BVSVEC),type,sizeof(type),&flg1);CHKERRQ(ierr);
     if (flg1) {
       ierr = BVSetType(bv,type);CHKERRQ(ierr);
     } else if (!((PetscObject)bv)->type_name) {
@@ -905,6 +912,9 @@ PetscErrorCode BVSetFromOptions(BV bv)
 
     ierr = PetscOptionsEnum("-bv_matmult","Method for BVMatMult","BVSetMatMultMethod",BVMatMultTypes,(PetscEnum)bv->vmm,(PetscEnum*)&bv->vmm,NULL);CHKERRQ(ierr);
 
+    ierr = PetscOptionsReal("-bv_definite_tol","Tolerance for checking a definite inner product","BVSetDefiniteTolerance",r,&r,&flg1);CHKERRQ(ierr);
+    if (flg1) { ierr = BVSetDefiniteTolerance(bv,r);CHKERRQ(ierr); }
+
     /* undocumented option to generate random vectors that are independent of the number of processes */
     ierr = PetscOptionsGetBool(NULL,NULL,"-bv_reproducible_random",&bv->rrandom,NULL);CHKERRQ(ierr);
 
@@ -914,9 +924,7 @@ PetscErrorCode BVSetFromOptions(BV bv)
     }
     ierr = PetscObjectProcessOptionsHandlers(PetscOptionsObject,(PetscObject)bv);CHKERRQ(ierr);
   ierr = PetscOptionsEnd();CHKERRQ(ierr);
-
-  if (!bv->rand) { ierr = BVGetRandomContext(bv,&bv->rand);CHKERRQ(ierr); }
-  ierr = PetscRandomSetFromOptions(bv->rand);CHKERRQ(ierr);
+  bv->sfocalled = PETSC_TRUE;
   PetscFunctionReturn(0);
 }
 
@@ -926,7 +934,7 @@ PetscErrorCode BVSetFromOptions(BV bv)
    for the block-orthogonalization (simultaneous orthogonalization of a set of
    vectors).
 
-   Logically Collective on BV
+   Logically Collective on bv
 
    Input Parameters:
 +  bv     - the basis vectors context
@@ -1034,7 +1042,7 @@ PetscErrorCode BVGetOrthogonalization(BV bv,BVOrthogType *type,BVOrthogRefineTyp
 /*@
    BVSetMatMultMethod - Specifies the method used for the BVMatMult() operation.
 
-   Logically Collective on BV
+   Logically Collective on bv
 
    Input Parameters:
 +  bv     - the basis vectors context
@@ -1046,7 +1054,7 @@ PetscErrorCode BVGetOrthogonalization(BV bv,BVOrthogType *type,BVOrthogRefineTyp
    Notes:
    Allowed values are
 +  BV_MATMULT_VECS - perform a matrix-vector multiply per each column
-.  BV_MATMULT_MAT - carry out a MatMatMult() product with a dense matrix
+.  BV_MATMULT_MAT - carry out a Mat-Mat product with a dense matrix
 -  BV_MATMULT_MAT_SAVE - this case is deprecated
 
    The default is BV_MATMULT_MAT except in the case of BVVECS.
@@ -1105,7 +1113,7 @@ PetscErrorCode BVGetMatMultMethod(BV bv,BVMatMultType *method)
    BVGetColumn - Returns a Vec object that contains the entries of the
    requested column of the basis vectors object.
 
-   Logically Collective on BV
+   Logically Collective on bv
 
    Input Parameters:
 +  bv - the basis vectors context
@@ -1139,6 +1147,7 @@ PetscErrorCode BVGetColumn(BV bv,PetscInt j,Vec *v)
   PetscValidHeaderSpecific(bv,BV_CLASSID,1);
   PetscValidType(bv,1);
   BVCheckSizes(bv,1);
+  BVCheckOp(bv,1,getcolumn);
   PetscValidLogicalCollectiveInt(bv,j,2);
   if (j<0 && -j>bv->nc) SETERRQ2(PetscObjectComm((PetscObject)bv),PETSC_ERR_ARG_OUTOFRANGE,"You requested constraint %D but only %D are available",-j,bv->nc);
   if (j>=bv->m) SETERRQ2(PetscObjectComm((PetscObject)bv),PETSC_ERR_ARG_OUTOFRANGE,"You requested column %D but only %D are available",j,bv->m);
@@ -1156,7 +1165,7 @@ PetscErrorCode BVGetColumn(BV bv,PetscInt j,Vec *v)
 /*@
    BVRestoreColumn - Restore a column obtained with BVGetColumn().
 
-   Logically Collective on BV
+   Logically Collective on bv
 
    Input Parameters:
 +  bv - the basis vectors context
@@ -1208,7 +1217,7 @@ PetscErrorCode BVRestoreColumn(BV bv,PetscInt j,Vec *v)
    BVGetArray - Returns a pointer to a contiguous array that contains this
    processor's portion of the BV data.
 
-   Logically Collective on BV
+   Logically Collective on bv
 
    Input Parameters:
 .  bv - the basis vectors context
@@ -1236,6 +1245,7 @@ PetscErrorCode BVGetArray(BV bv,PetscScalar **a)
   PetscValidHeaderSpecific(bv,BV_CLASSID,1);
   PetscValidType(bv,1);
   BVCheckSizes(bv,1);
+  BVCheckOp(bv,1,getarray);
   ierr = (*bv->ops->getarray)(bv,a);CHKERRQ(ierr);
   PetscFunctionReturn(0);
 }
@@ -1243,7 +1253,7 @@ PetscErrorCode BVGetArray(BV bv,PetscScalar **a)
 /*@C
    BVRestoreArray - Restore the BV object after BVGetArray() has been called.
 
-   Logically Collective on BV
+   Logically Collective on bv
 
    Input Parameters:
 +  bv - the basis vectors context
@@ -1305,6 +1315,7 @@ PetscErrorCode BVGetArrayRead(BV bv,const PetscScalar **a)
   PetscValidHeaderSpecific(bv,BV_CLASSID,1);
   PetscValidType(bv,1);
   BVCheckSizes(bv,1);
+  BVCheckOp(bv,1,getarrayread);
   ierr = (*bv->ops->getarrayread)(bv,a);CHKERRQ(ierr);
   PetscFunctionReturn(0);
 }
@@ -1342,7 +1353,7 @@ PetscErrorCode BVRestoreArrayRead(BV bv,const PetscScalar **a)
    BVCreateVec - Creates a new Vec object with the same type and dimensions
    as the columns of the basis vectors object.
 
-   Collective on BV
+   Collective on bv
 
    Input Parameter:
 .  bv - the basis vectors context
@@ -1373,7 +1384,7 @@ PetscErrorCode BVCreateVec(BV bv,Vec *v)
    BVCreateMat - Creates a new Mat object of dense type and copies the contents
    of the BV object.
 
-   Collective on BV
+   Collective on bv
 
    Input Parameter:
 .  bv - the basis vectors context
@@ -1402,11 +1413,11 @@ PetscErrorCode BVCreateMat(BV bv,Mat *A)
   PetscValidPointer(A,2);
 
   ierr = MatCreateDense(PetscObjectComm((PetscObject)bv->t),bv->n,PETSC_DECIDE,bv->N,bv->m,NULL,A);CHKERRQ(ierr);
-  ierr = MatDenseGetArray(*A,&aa);CHKERRQ(ierr);
+  ierr = MatDenseGetArrayWrite(*A,&aa);CHKERRQ(ierr);
   ierr = BVGetArrayRead(bv,&vv);CHKERRQ(ierr);
-  ierr = PetscMemcpy(aa,vv,bv->m*bv->n*sizeof(PetscScalar));CHKERRQ(ierr);
+  ierr = PetscArraycpy(aa,vv,bv->m*bv->n);CHKERRQ(ierr);
   ierr = BVRestoreArrayRead(bv,&vv);CHKERRQ(ierr);
-  ierr = MatDenseRestoreArray(*A,&aa);CHKERRQ(ierr);
+  ierr = MatDenseRestoreArrayWrite(*A,&aa);CHKERRQ(ierr);
   ierr = MatAssemblyBegin(*A,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
   ierr = MatAssemblyEnd(*A,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
   PetscFunctionReturn(0);
@@ -1416,7 +1427,7 @@ PetscErrorCode BVCreateMat(BV bv,Mat *A)
    BVGetMat - Returns a Mat object of dense type that shares the memory of
    the BV object.
 
-   Collective on BV
+   Collective on bv
 
    Input Parameter:
 .  bv - the basis vectors context
@@ -1447,34 +1458,36 @@ PetscErrorCode BVGetMat(BV bv,Mat *A)
   PetscValidHeaderSpecific(bv,BV_CLASSID,1);
   BVCheckSizes(bv,1);
   PetscValidPointer(A,2);
-  m = bv->k-bv->l;
-  if (!bv->Aget) create=PETSC_TRUE;
-  else {
-    ierr = MatDenseGetArray(bv->Aget,&aa);CHKERRQ(ierr);
-    if (aa) SETERRQ(PetscObjectComm((PetscObject)bv),PETSC_ERR_ARG_WRONGSTATE,"BVGetMat already called on this BV");
-    ierr = MatGetSize(bv->Aget,NULL,&cols);CHKERRQ(ierr);
-    if (cols!=m) {
-      ierr = MatDestroy(&bv->Aget);CHKERRQ(ierr);
-      create=PETSC_TRUE;
+  if (bv->ops->getmat) {
+    ierr = (*bv->ops->getmat)(bv,A);CHKERRQ(ierr);
+  } else {
+    m = bv->k-bv->l;
+    if (!bv->Aget) create=PETSC_TRUE;
+    else {
+      ierr = MatDenseGetArray(bv->Aget,&aa);CHKERRQ(ierr);
+      if (aa) SETERRQ(PetscObjectComm((PetscObject)bv),PETSC_ERR_ARG_WRONGSTATE,"BVGetMat already called on this BV");
+      ierr = MatGetSize(bv->Aget,NULL,&cols);CHKERRQ(ierr);
+      if (cols!=m) {
+        ierr = MatDestroy(&bv->Aget);CHKERRQ(ierr);
+        create=PETSC_TRUE;
+      }
     }
+    ierr = BVGetArray(bv,&vv);CHKERRQ(ierr);
+    if (create) {
+      ierr = MatCreateDense(PetscObjectComm((PetscObject)bv),bv->n,PETSC_DECIDE,bv->N,m,vv,&bv->Aget);CHKERRQ(ierr); /* pass a pointer to avoid allocation of storage */
+      ierr = MatDenseReplaceArray(bv->Aget,NULL);CHKERRQ(ierr);  /* replace with a null pointer, the value after BVRestoreMat */
+      ierr = PetscLogObjectParent((PetscObject)bv,(PetscObject)bv->Aget);CHKERRQ(ierr);
+    }
+    ierr = MatDensePlaceArray(bv->Aget,vv+(bv->nc+bv->l)*bv->n);CHKERRQ(ierr);  /* set the actual pointer */
+    *A = bv->Aget;
   }
-  ierr = BVGetArray(bv,&vv);CHKERRQ(ierr);
-  if (create) {
-    ierr = MatCreateDense(PetscObjectComm((PetscObject)bv),bv->n,PETSC_DECIDE,bv->N,m,vv,&bv->Aget);CHKERRQ(ierr); /* pass a pointer to avoid allocation of storage */
-    ierr = MatDensePlaceArray(bv->Aget,NULL);CHKERRQ(ierr);  /* replace with a null pointer, the value after BVRestoreMat */
-    ierr = PetscLogObjectParent((PetscObject)bv,(PetscObject)bv->Aget);CHKERRQ(ierr);
-    ierr = MatAssemblyBegin(bv->Aget,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
-    ierr = MatAssemblyEnd(bv->Aget,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
-  }
-  ierr = MatDensePlaceArray(bv->Aget,vv+(bv->nc+bv->l)*bv->n);CHKERRQ(ierr);  /* set the actual pointer */
-  *A = bv->Aget;
   PetscFunctionReturn(0);
 }
 
 /*@
    BVRestoreMat - Restores the Mat obtained with BVGetMat().
 
-   Collective on BV
+   Logically Collective on bv
 
    Input Parameters:
 +  bv - the basis vectors context
@@ -1500,11 +1513,15 @@ PetscErrorCode BVRestoreMat(BV bv,Mat *A)
   PetscValidPointer(A,2);
   if (!bv->Aget) SETERRQ(PetscObjectComm((PetscObject)bv),PETSC_ERR_ARG_WRONGSTATE,"BVRestoreMat must match a previous call to BVGetMat");
   if (bv->Aget!=*A) SETERRQ(PetscObjectComm((PetscObject)bv),PETSC_ERR_ARG_WRONGSTATE,"Mat argument is not the same as the one obtained with BVGetMat");
-  ierr = MatDenseGetArray(bv->Aget,&aa);CHKERRQ(ierr);
-  vv = aa-(bv->nc+bv->l)*bv->n;
-  ierr = MatDenseResetArray(bv->Aget);CHKERRQ(ierr);
-  ierr = BVRestoreArray(bv,&vv);CHKERRQ(ierr);
-  *A = NULL;
+  if (bv->ops->restoremat) {
+    ierr = (*bv->ops->restoremat)(bv,A);CHKERRQ(ierr);
+  } else {
+    ierr = MatDenseGetArray(bv->Aget,&aa);CHKERRQ(ierr);
+    vv = aa-(bv->nc+bv->l)*bv->n;
+    ierr = MatDenseResetArray(bv->Aget);CHKERRQ(ierr);
+    ierr = BVRestoreArray(bv,&vv);CHKERRQ(ierr);
+    *A = NULL;
+  }
   PetscFunctionReturn(0);
 }
 
@@ -1521,11 +1538,15 @@ PETSC_STATIC_INLINE PetscErrorCode BVDuplicate_Private(BV V,BV W)
   W->orthog_ref   = V->orthog_ref;
   W->orthog_eta   = V->orthog_eta;
   W->orthog_block = V->orthog_block;
-  if (V->matrix) PetscObjectReference((PetscObject)V->matrix);
+  if (V->matrix) { ierr = PetscObjectReference((PetscObject)V->matrix);CHKERRQ(ierr); }
   W->matrix       = V->matrix;
   W->indef        = V->indef;
   W->vmm          = V->vmm;
   W->rrandom      = V->rrandom;
+  W->deftol       = V->deftol;
+  if (V->rand) { ierr = PetscObjectReference((PetscObject)V->rand);CHKERRQ(ierr); }
+  W->rand         = V->rand;
+  W->sfocalled    = V->sfocalled;
   if (V->ops->duplicate) { ierr = (*V->ops->duplicate)(V,W);CHKERRQ(ierr); }
   ierr = PetscObjectStateIncrease((PetscObject)W);CHKERRQ(ierr);
   PetscFunctionReturn(0);
@@ -1535,7 +1556,7 @@ PETSC_STATIC_INLINE PetscErrorCode BVDuplicate_Private(BV V,BV W)
    BVDuplicate - Creates a new basis vector object of the same type and
    dimensions as an existing one.
 
-   Collective on BV
+   Collective on V
 
    Input Parameter:
 .  V - basis vectors context
@@ -1574,7 +1595,7 @@ PetscErrorCode BVDuplicate(BV V,BV *W)
    BVDuplicateResize - Creates a new basis vector object of the same type and
    dimensions as an existing one, but with possibly different number of columns.
 
-   Collective on BV
+   Collective on V
 
    Input Parameter:
 +  V - basis vectors context
@@ -1647,7 +1668,7 @@ PetscErrorCode BVGetCachedBV(BV bv,BV *cached)
 /*@
    BVCopy - Copies a basis vector object into another one, W <- V.
 
-   Logically Collective on BV
+   Logically Collective on V
 
    Input Parameter:
 .  V - basis vectors context
@@ -1675,13 +1696,14 @@ PetscErrorCode BVCopy(BV V,BV W)
   PetscValidHeaderSpecific(V,BV_CLASSID,1);
   PetscValidType(V,1);
   BVCheckSizes(V,1);
+  BVCheckOp(V,1,copy);
   PetscValidHeaderSpecific(W,BV_CLASSID,2);
   PetscValidType(W,2);
   BVCheckSizes(W,2);
   PetscCheckSameTypeAndComm(V,1,W,2);
   if (V->n!=W->n) SETERRQ2(PetscObjectComm((PetscObject)V),PETSC_ERR_ARG_INCOMP,"Mismatching local dimension V %D, W %D",V->n,W->n);
   if (V->k-V->l>W->m-W->l) SETERRQ2(PetscObjectComm((PetscObject)V),PETSC_ERR_ARG_SIZ,"W has %D non-leading columns, not enough to store %D columns",W->m-W->l,V->k-V->l);
-  if (!V->n) PetscFunctionReturn(0);
+  if (V==W || !V->n) PetscFunctionReturn(0);
 
   ierr = PetscLogEventBegin(BV_Copy,V,W,0,0);CHKERRQ(ierr);
   if (V->indef && V->matrix && V->indef==W->indef && V->matrix==W->matrix) {
@@ -1689,7 +1711,7 @@ PetscErrorCode BVCopy(BV V,BV W)
     ierr = BV_AllocateSignature(W);CHKERRQ(ierr);
     ierr = VecGetArrayRead(V->omega,&vomega);CHKERRQ(ierr);
     ierr = VecGetArray(W->omega,&womega);CHKERRQ(ierr);
-    ierr = PetscMemcpy(womega+W->nc+W->l,vomega+V->nc+V->l,(V->k-V->l)*sizeof(PetscScalar));CHKERRQ(ierr);
+    ierr = PetscArraycpy(womega+W->nc+W->l,vomega+V->nc+V->l,V->k-V->l);CHKERRQ(ierr);
     ierr = VecRestoreArray(W->omega,&womega);CHKERRQ(ierr);
     ierr = VecRestoreArrayRead(V->omega,&vomega);CHKERRQ(ierr);
   }
@@ -1702,7 +1724,7 @@ PetscErrorCode BVCopy(BV V,BV W)
 /*@
    BVCopyVec - Copies one of the columns of a basis vectors object into a Vec.
 
-   Logically Collective on BV
+   Logically Collective on V
 
    Input Parameter:
 +  V - basis vectors context
@@ -1747,7 +1769,7 @@ PetscErrorCode BVCopyVec(BV V,PetscInt j,Vec w)
 /*@
    BVCopyColumn - Copies the values from one of the columns to another one.
 
-   Logically Collective on BV
+   Logically Collective on V
 
    Input Parameter:
 +  V - basis vectors context
@@ -1799,6 +1821,7 @@ static PetscErrorCode BVGetSplit_Private(BV bv,PetscBool left,BV *split)
 
   PetscFunctionBegin;
   ncols = left? bv->nc+bv->l: bv->m-bv->l;
+  if (*split && ncols!=(*split)->m) { ierr = BVDestroy(split);CHKERRQ(ierr); }
   if (!*split) {
     ierr = BVCreate(PetscObjectComm((PetscObject)bv),split);CHKERRQ(ierr);
     ierr = PetscLogObjectParent((PetscObject)bv,(PetscObject)*split);CHKERRQ(ierr);
@@ -1806,8 +1829,6 @@ static PetscErrorCode BVGetSplit_Private(BV bv,PetscBool left,BV *split)
     (*split)->splitparent = bv;
     ierr = BVSetSizesFromVec(*split,bv->t,ncols);CHKERRQ(ierr);
     ierr = BVDuplicate_Private(bv,*split);CHKERRQ(ierr);
-  } else {
-    ierr = BVResize(*split,ncols,PETSC_FALSE);CHKERRQ(ierr);
   }
   (*split)->l  = 0;
   (*split)->k  = left? bv->l: bv->k-bv->l;
@@ -1830,7 +1851,7 @@ static PetscErrorCode BVGetSplit_Private(BV bv,PetscBool left,BV *split)
    internal data, one of them containing the leading columns and the other
    one containing the remaining columns.
 
-   Logically Collective on BV
+   Logically Collective on bv
 
    Input Parameters:
 .  bv - the basis vectors context
@@ -1878,7 +1899,7 @@ PetscErrorCode BVGetSplit(BV bv,BV *L,BV *R)
 /*@
    BVRestoreSplit - Restore the BV objects obtained with BVGetSplit().
 
-   Logically Collective on BV
+   Logically Collective on bv
 
    Input Parameters:
 +  bv - the basis vectors context
@@ -1915,3 +1936,70 @@ PetscErrorCode BVRestoreSplit(BV bv,BV *L,BV *R)
   PetscFunctionReturn(0);
 }
 
+/*@
+   BVSetDefiniteTolerance - Set the tolerance to be used when checking a
+   definite inner product.
+
+   Logically Collective on bv
+
+   Input Parameters:
++  bv     - basis vectors
+-  deftol - the tolerance
+
+   Options Database Key:
+.  -bv_definite_tol <deftol> - the tolerance
+
+   Notes:
+   When using a non-standard inner product, see BVSetMatrix(), the solver needs
+   to compute sqrt(z'*B*z) for various vectors z. If the inner product has not
+   been declared indefinite, the value z'*B*z must be positive, but due to
+   rounding error a tiny value may become negative. A tolerance is used to
+   detect this situation. Likewise, in complex arithmetic z'*B*z should be
+   real, and we use the same tolerance to check whether a nonzero imaginary part
+   can be considered negligible.
+
+   This function sets this tolerance, which defaults to 10*eps, where eps is
+   the machine epsilon. The default value should be good for most applications.
+
+   Level: advanced
+
+.seealso: BVSetMatrix()
+@*/
+PetscErrorCode BVSetDefiniteTolerance(BV bv,PetscReal deftol)
+{
+  PetscFunctionBegin;
+  PetscValidHeaderSpecific(bv,BV_CLASSID,1);
+  PetscValidLogicalCollectiveReal(bv,deftol,2);
+  if (deftol == PETSC_DEFAULT) bv->deftol = 10*PETSC_MACHINE_EPSILON;
+  else {
+    if (deftol<=0.0) SETERRQ(PetscObjectComm((PetscObject)bv),PETSC_ERR_ARG_OUTOFRANGE,"Illegal value of deftol. Must be > 0");
+    bv->deftol = deftol;
+  }
+  PetscFunctionReturn(0);
+}
+
+/*@
+   BVGetDefiniteTolerance - Returns the tolerance for checking a definite
+   inner product.
+
+   Not Collective
+
+   Input Parameter:
+.  bv - the basis vectors
+
+   Output Parameters:
+.  deftol - the tolerance
+
+   Level: advanced
+
+.seealso: BVGetDefiniteTolerance()
+@*/
+PetscErrorCode BVGetDefiniteTolerance(BV bv,PetscReal *deftol)
+{
+  PetscFunctionBegin;
+  PetscValidHeaderSpecific(bv,BV_CLASSID,1);
+  PetscValidRealPointer(deftol,2);
+  *deftol = bv->deftol;
+  PetscFunctionReturn(0);
+}
+
